unit GTOPO30toHeightForm;

{
Revision history:

V1.0.0  2000 Apr 08  First version
V1.1.0  2000 Apr 26  Move to fresh folder
                     Rename to GTOPO30toHeightImg.dpr
V1.2.0  2000 May 22  Make grey scale range nearer full dynamic range
                     Add Lakes data from GSHHS file
V1.3.0  2000 May 27  Add Close button
                     Save file/folder locations across invocations
V1.3.2  2000 May 28  Noted that extreme west of Antartic tile appears blank
                     So you may need to edit the PNG file....
V1.4.0  2001 Sep 21  Version for Delphi 5
                     Alter height scaling to 3 * sqrt (metres)
                     Write to files colours.png and topo.png
                     Uses PNG_IO.pas
V1.5.0  2002 Jan 24  Add LandSeaMask.png (mask file)
                     Make default subsample 8
                     Remove depandancy on TExtBitmap in BMPExtImage unit

Copyright  David J Taylor, Edinburgh
Web site:  www.satsignal.net
E-mail:    davidtaylor@writeme.com
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls, TileData, StdCtrls, RegistryVariables,
  ComCtrls;

const
  // 8, 12, 24 and 48 may also be acceptable, depending on your display driver!
  subsample = 8;  // was 8 or 12
  colour = False;

type
  TFormMain = class (TForm)
    MainMenu1: TMainMenu;
    FileMenu: TMenuItem;
    FileExit: TMenuItem;
    HelpMenu: TMenuItem;
    HelpAbout: TMenuItem;
    ImageDisplay: TImage;
    ButtonRefresh: TButton;
    ButtonClose: TButton;
    OpenDialog: TOpenDialog;
    ProgressBar: TProgressBar;
    procedure FormCreate (Sender: TObject);
    procedure FormDestroy (Sender: TObject);
    procedure FileExitClick (Sender: TObject);
    procedure HelpAboutClick (Sender: TObject);
    procedure ButtonRefreshClick (Sender: TObject);
    procedure ButtonCloseClick (Sender: TObject);
  private
    { Private declarations }
    FBMP: TBitmap;
    FLandSeaMask: TBitmap;
    Tiles: TTileArray;
    GSHHS_file: TRegistryString;    // Use a string variable as an alternative
    GTOPO30_folder: TRegistryString;
    procedure add_lake_data;
    procedure make_greyscale (bmp: TBitmap);
  public
    { Public declarations }
  end;

var
  FormMain: TFormMain;


implementation

uses
  AboutDisplay, ElevationColours, GSHHSdata, PNG_IO;

{$R *.DFM}

procedure TFormMain.FormCreate (Sender: TObject);
begin
  ImageDisplay.Picture.Bitmap := TBitmap.Create;
  FBmp := ImageDisplay.Picture.Bitmap;
  with FBmp do
    begin
    if colour
      then PixelFormat := pf24bit
      else PixelFormat := pf8bit;
    Width := array_cols div subsample;
    Height := array_rows div subsample;
    make_greyscale (FBmp);
    end;

  FLandSeaMask := TBitmap.Create;
  with FLandSeaMask do
    begin
    PixelFormat := pf8bit;
    Width := array_cols div subsample;
    Height := array_rows div subsample;
    make_greyscale (FLandSeaMask);
    end;

  GTOPO30_folder := TRegistryString.CreateValue ('GTOPO30Folder', 'C:\GTOPO30\');
  // Use the Low-res data (5 km)
  if not FileExists (GTOPO30_folder.Value + 'E020N40.dem') then
    with OpenDialog do
      begin
      FileName := 'E020N40.dem';
      Title := 'Please point to the folder containing the GTOPO30 data';
      if Execute
        then GTOPO30_folder.Value := ExtractFilePath (FileName)
        else
          begin
          ShowMessage ('GTOPO30 Height to Image cannot continue without GTOPO30 data files!');
          Application.Terminate;
          end;
      end;
  Tiles := TTileArray.Create (GTOPO30_folder.Value);
  GSHHS_file := TRegistryString.CreateValue ('GSHHSFile', 'C:\GSHHS\gshhs_l.b');
  if not FileExists (GSHHS_file.Value) then
    with OpenDialog do
      begin
      FileName := 'GSHHS*.*';
      InitialDir := 'C:\';
      Title := 'Please point to the GSHHS file you would like to use';
      if Execute
        then GSHHS_file.Value := FileName;
      end;
  if not FileExists (GSHHS_file.Value) then
    begin
    ShowMessage ('GTOPO30 Height to Image cannot continue without a GSHHS data file!');
    Application.Terminate;
    end;
end;


procedure TFormMain.FormDestroy (Sender: TObject);
begin
  if Assigned (Tiles) then Tiles.Free;
  GTOPO30_folder.Free;
  GSHHS_file.Free;
end;


procedure TFormMain.FileExitClick (Sender: TObject);
begin
  Close;
end;


procedure TFormMain.HelpAboutClick (Sender: TObject);
begin
  AboutBoxDJT1.ShowModal;
end;


procedure TFormMain.ButtonRefreshClick (Sender: TObject);
var
  col, row: integer;
  x, y: integer;
  c: TColor;
  pt: ^TRGBTriple;
  pb: ^Byte;
  pm: ^Byte;
  tile_col, tile_row: integer;
  max_tile_col: integer;
  y_base, dx, dy: integer;
  h, i: integer;
  saved_cursor: TCursor;
  filename: string;
  height_to_grey: array [SmallInt] of byte;
begin
  if not Assigned (Tiles) then Exit;

  saved_cursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  ProgressBar.Max := 27;

  // Mount Everst is 8850m high which would be 282 on the following mapping
  for h := -32768 to +32767 do
    begin
    height_to_grey [h] := 0;
    if h <> -9999 then
      begin
      if h < 1
      then i := 1
      else
        begin
        i := Round (3 * Sqrt (h));
        end;
      if i > 255 then i := 255;
      height_to_grey [h] := i;
      end;
    end;

  with FBmp do
    begin
    y_base := Height * 50 div 180;
    for tile_row := 0 to 3 do
      begin
      case tile_row of
        3: begin
           dx := Width * 60 div 360;
           dy := Height * 30 div 180;
           max_tile_col := 6;
           end;
      else
        dx := Width * 40 div 360;
        dy := Height * 50 div 180;
        max_tile_col := 9;
      end;
      for tile_col := 0 to max_tile_col - 1 do
        begin
        ProgressBar.Position := ProgressBar.Position + 1;
        for row := tile_row * y_base to tile_row * y_base + dy - 1 do
          begin
          y := row * array_rows div FBmp.Height;
          pt := ScanLine [row];       // For a colour image
          pb := ScanLine [row];       // For a monochrome one
          pm := FLandSeaMask.ScanLine [row];
          Inc (pt, tile_col * dx);
          Inc (pb, tile_col * dx);
          Inc (pm, tile_col * dx);
          for col := tile_col * dx to (tile_col + 1) * dx - 1 do
            begin
            x := col * array_cols div Width;
            h := Tiles.HeightAt (x, y);

            if colour
            then
              begin
              c := colours [h];
              pt^.rgbtRed := c and $ff;
              pt^.rgbtGreen := (c and $FF00) shr 8;
              pt^.rgbtBlue := (c and $FF0000) shr 16;
              Inc (pt);
              end
            else
              begin
              pb^ := height_to_grey [h];
              Inc (pb);
              if h > 0
                then pm^ := 255
                else pm^ := 0;
              Inc (pm);
              end;
            end;
          end;
        Modified := True;
        ImageDisplay.Update;
        end;
      end;
    with FBmp do
      begin
      Canvas.Pen.Color := clWhite;
      Canvas.Pen.Style := psClear;
      Canvas.Brush.Color := clBlack;
      end;
    with FLandSeaMask do
      begin
      Canvas.Pen.Color := clWhite;
      Canvas.Pen.Style := psClear;
      Canvas.Brush.Color := clBlack;
      end;
    add_lake_data;
    Modified := True;
    ImageDisplay.Update;
    end;

  FLandSeaMask.Modified := True;

  ProgressBar.Position := 0;
  if colour
    then filename := 'colours.png'
    else filename := 'topo.png';
  SaveBmpAsPngFile (FBmp, filename,
                    'David Taylor http://www.satsignal.net',
                    'Scaled World Map Heights',
                    'GTOPO30toHeightImg',
                    'Topographic Data');
  SaveBmpAsPngFile (FLandSeaMask, 'LandSeaMask.png',
                    'David Taylor http://www.satsignal.net',
                    'Scaled World Map Mask',
                    'GTOPO30toHeightImg',
                    'Topographic Data');
  ShowMessage ('Image(s) written to file ' + filename);
  Screen.Cursor := saved_cursor;
end;


procedure TFormMain.add_lake_data;

  procedure read_one_line (const hdr: TGSHHS;  var f: file);
  type
    point_array = array [0..0] of TPoint;
  var
    npoints: integer;
    point: integer;
    old_x, old_y: integer;
    xy: TGPOINT;
    bytes_read: integer;
    ax, ay: single;
    x, y: integer;
    ppoints: ^point_array;
    active_points: integer;
  begin
    npoints := hdr.n;
    GetMem (ppoints, npoints * SizeOf (TPoint));
    old_x := -1;
    old_y := -1;
    active_points := 0;
    with FBmp do
      begin
      for point := 1 to npoints do
        begin
        BlockRead (f, xy, SizeOf (xy), bytes_read);
        if bytes_read = SizeOf (xy) then
          begin
          PointSwap (xy);
          ax := xy.x * 1E-6;
          ax := ax + 180;
          while ax >= 360 do ax := ax - 360;
          x := Round (ax * Width / 360);
          ay := xy.y * 1E-6;
          y := Round ((90 - ay) * Height / 180);
          if point = 1
          then                // First point, simply move there
            begin
            ppoints^ [active_points].X := x;
            ppoints^ [active_points].Y := y;
            old_x := x;
            old_y := y;
            Inc (active_points);
            end
          else                  // There is a genuine move, or the last point
            if (x <> old_x) or (y <> old_y) or (point = npoints) then
              begin
            ppoints^ [active_points].X := x;
            ppoints^ [active_points].Y := y;
                old_x := x;     // Note the new pen co-ordinates
                old_y := y;
                Inc (active_points);
              end;
          end;
        end;
      Canvas.Polygon (Slice (ppoints^, active_points));
      FLandSeaMask.Canvas.Polygon (Slice (ppoints^, active_points));
      end;
    FreeMem (ppoints);
    end;

var
  f: file;
  hdr: TGSHHS;
  eof: boolean;
  bytes_read: integer;
begin
  try
    AssignFile (f, GSHHS_file.Value);
    Reset (f, 1);
    repeat
      BlockRead (f, hdr, SizeOf (hdr), bytes_read);
      eof := bytes_read <> SizeOf (hdr);
      if not eof then
        begin
        HeaderSwap (hdr);
        // Only include lakes, and then only sizeable ones
        // Area to exceed 20000 of 1/10 km^2
        if (hdr.level = lvLake) and (hdr.area > 20000)
          then read_one_line (hdr, f)
          else Seek (f, FilePos (f) + hdr.n * SizeOf (TGPoint));
        end;
    until eof;
    CloseFile (f);
  except
  end;
end;


procedure TFormMain.ButtonCloseClick (Sender: TObject);
begin
  Close;
end;

procedure TFormMain.make_greyscale (bmp: TBitmap);
var
  b: integer;
  log_pal: TMaxLogPalette;
  pal: HPalette;
begin
  with bmp do
    begin
    log_pal.palVersion := $0300;
    log_pal.palNumEntries := 256;
    for b := 0 to 255 do
      with log_pal.palPalEntry [b] do
        begin
        peBlue := b;
        peGreen := b;
        peRed := b;
        peFlags := 0;
        end;
    pal := CreatePalette (PLogPalette(@log_pal)^);
    Palette := pal;
    Modified := True;
    end;
end;

end.


